
BEGIN {
  unless ($ENV{RELEASE_TESTING}) {
    require Test::More;
    Test::More::plan(skip_all => 'these tests are for release candidate testing');
  }
}

use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::NoTabs 0.08

use Test::More 0.88;
use Test::NoTabs;

my @files = (
    'bin/treex',
    'bin/ttred',
    'lib/Treex.pm',
    'lib/Treex/Block/Read/BaseReader.pm',
    'lib/Treex/Block/Read/BaseTextReader.pm',
    'lib/Treex/Block/Read/Sentences.pm',
    'lib/Treex/Block/Read/Text.pm',
    'lib/Treex/Block/Read/Treex.pm',
    'lib/Treex/Block/Util/DefinedAttr.pm',
    'lib/Treex/Block/Util/Eval.pm',
    'lib/Treex/Block/Util/Find.pm',
    'lib/Treex/Block/Util/SetGlobal.pm',
    'lib/Treex/Block/Write/BaseTextWriter.pm',
    'lib/Treex/Block/Write/BaseWriter.pm',
    'lib/Treex/Block/Write/Sentences.pm',
    'lib/Treex/Block/Write/Text.pm',
    'lib/Treex/Block/Write/Treex.pm',
    'lib/Treex/Core.pm',
    'lib/Treex/Core/Block.pm',
    'lib/Treex/Core/Bundle.pm',
    'lib/Treex/Core/BundleZone.pm',
    'lib/Treex/Core/CacheBlock.pm',
    'lib/Treex/Core/Cloud.pm',
    'lib/Treex/Core/Common.pm',
    'lib/Treex/Core/Config.pm',
    'lib/Treex/Core/Coordination.pm',
    'lib/Treex/Core/DocZone.pm',
    'lib/Treex/Core/Document.pm',
    'lib/Treex/Core/DocumentReader.pm',
    'lib/Treex/Core/DocumentReader/Base.pm',
    'lib/Treex/Core/DocumentReader/ZoneReader.pm',
    'lib/Treex/Core/Files.pm',
    'lib/Treex/Core/Loader.pm',
    'lib/Treex/Core/Log.pm',
    'lib/Treex/Core/Node.pm',
    'lib/Treex/Core/Node/A.pm',
    'lib/Treex/Core/Node/EffectiveRelations.pm',
    'lib/Treex/Core/Node/InClause.pm',
    'lib/Treex/Core/Node/Interset.pm',
    'lib/Treex/Core/Node/N.pm',
    'lib/Treex/Core/Node/Ordered.pm',
    'lib/Treex/Core/Node/P.pm',
    'lib/Treex/Core/Node/T.pm',
    'lib/Treex/Core/Phrase.pm',
    'lib/Treex/Core/Phrase/BaseNTerm.pm',
    'lib/Treex/Core/Phrase/Builder.pm',
    'lib/Treex/Core/Phrase/Coordination.pm',
    'lib/Treex/Core/Phrase/NTerm.pm',
    'lib/Treex/Core/Phrase/PP.pm',
    'lib/Treex/Core/Phrase/Term.pm',
    'lib/Treex/Core/RememberArgs.pm',
    'lib/Treex/Core/Resource.pm',
    'lib/Treex/Core/Run.pm',
    'lib/Treex/Core/Scenario.pm',
    'lib/Treex/Core/ScenarioParser.pm',
    'lib/Treex/Core/TredView.pm',
    'lib/Treex/Core/TredView/AnnotationCommand.pm',
    'lib/Treex/Core/TredView/BackendStorable.pm',
    'lib/Treex/Core/TredView/Colors.pm',
    'lib/Treex/Core/TredView/Common.pm',
    'lib/Treex/Core/TredView/Labels.pm',
    'lib/Treex/Core/TredView/LineStyles.pm',
    'lib/Treex/Core/TredView/Styles.pm',
    'lib/Treex/Core/TredView/TreeLayout.pm',
    'lib/Treex/Core/TredView/Vallex.pm',
    'lib/Treex/Core/Types.pm',
    'lib/Treex/Core/WildAttr.pm',
    'lib/Treex/Core/Zone.pm',
    'lib/Treex/Tool/Probe.pm',
    't/00-compile.t',
    't/a_effective_rel.t',
    't/alignment.t',
    't/attribute.t',
    't/author-critic.t',
    't/backrefs.t',
    't/base.t',
    't/block.t',
    't/bundle.t',
    't/bundle_zones.t',
    't/common.t',
    't/config.t',
    't/document.t',
    't/document_zones.t',
    't/dump_scen.t',
    't/error.scen',
    't/files.t',
    't/following_node.t',
    't/grammar.t',
    't/log.t',
    't/n_tree.t',
    't/node.t',
    't/ordered.t',
    't/p_penn_treebank_string_format.t',
    't/phrase.t',
    't/release-cpan-changes.t',
    't/release-dist-manifest.t',
    't/release-distmeta.t',
    't/release-kwalitee.t',
    't/release-minimum-version.t',
    't/release-mojibake.t',
    't/release-no-tabs.t',
    't/release-pod-coverage.t',
    't/release-pod-linkcheck.t',
    't/release-pod-syntax.t',
    't/release-portability.t',
    't/release-synopsis.t',
    't/release-test-version.t',
    't/release-unused-vars.t',
    't/remove_bundle.t',
    't/resource.t',
    't/run.t',
    't/runall.sh',
    't/scenario.t',
    't/scenario_parser.t',
    't/storable.t',
    't/t_effective_rel.t',
    't/test.scen',
    't/trees.t',
    't/wildattr.t',
    't/writers.t'
);

notabs_ok($_) foreach @files;
done_testing;
